use crate::transformer::Transformers;
use crate::point::Point;
use crate::worldstate::{WorldState,EDH};
use serde::{Serialize,Deserialize};
use log::debug;
use nalgebra::Matrix1x3;

/// Homography

#[derive(Serialize,Deserialize,Debug,Clone,Copy)]
pub struct Homography {}

impl Transformers for Homography {
    fn apply(&self, point_list: &[Point], ws: &WorldState) -> Vec<Point> {	
	let edh : &EDH = &ws.edh;

	point_list.iter()
	    .map(| point | {
		let p = Matrix1x3::new(point.x, point.y, 1.0);
		let p = p * edh.matrix;
		let new_p = Point { x: p[0] / p[2], y: p[1] / p[2], ..*point };
		
		debug!("{:?} => {:?}", point, new_p);
		
		new_p
	    })
	    .collect()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_homography_identity() {
	let p0 = Point { x: -1500.0, y: 1500.0, ..Point::default() };

	let edh = EDH::new(vec![
	    vec![ 1.0, 0.0, 0.0 ],
	    vec![ 0.0, 1.0, 0.0 ],
	    vec![ 0.0, 0.0, 1.0 ]
	]).unwrap();

	let ws = WorldState { edh : edh, ..WorldState::default() };
	
	let homography = Homography{};
	let result = homography.apply(&[p0], &ws);
	
	assert_eq!(result, vec![Point { x: -1500.0,
					y: 1500.0,
					..Point::default() }]);
    }

    #[test]
    fn test_homography_rotation() {
	let p0 = Point { x: -1500.0, y: 1500.0, ..Point::default() };

	let edh = EDH::new(vec![
	    vec![ 1.24107321e-03, 1.00500127e-03, 7.15439347e-01],
	    vec![-9.93223912e-04, 1.22652939e-03,-6.98671238e-01],
	    vec![ 1.06017142e-17,-4.69459541e-17, 3.32700590e-05]
	]).unwrap();

	let ws = WorldState { edh : edh, ..WorldState::default() };
	
	let homography = Homography{};
	let result = homography.apply(&[p0], &ws);
	
	assert_eq!(result, vec![Point { x: 10860.557,
					y: 79078.87,
					..Point::default() }]);
    }
}
